# -*- coding: utf-8 -*-
import os
import configparser

from kinto import main as kinto_main
from kinto.core.testing import get_user_headers, BaseWebTest as CoreWebTest


here = os.path.abspath(os.path.dirname(__file__))


class BaseWebTest(CoreWebTest):
    api_prefix = "v1"
    entry_point = kinto_main
    config = 'config.ini'

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.headers.update(get_user_headers('mat'))
        cls.indexer = cls.app.app.registry.indexer

    def tearDown(self):
        super().tearDown()
        self.indexer.join()
        self.indexer.flush()

    @classmethod
    def ini_path(cls):
        ini_local_path = os.path.join(here, 'config_local.ini')
        ini_path = os.path.join(here, 'config.ini')
        if os.path.exists(ini_local_path):
            ini_path = ini_local_path
        return ini_path

    @classmethod
    def get_app_settings(cls, extras=None):
        config = configparser.ConfigParser()
        config.read(cls.ini_path())
        settings = dict(config.items('app:main'))
        if extras:
            settings.update(extras)
        return settings



import logging
from copy import deepcopy
from contextlib import contextmanager

from algoliasearch.http.verb import Verb
from algoliasearch.search_client import SearchClient
from algoliasearch.exceptions import AlgoliaException
from pyramid.exceptions import ConfigurationError


logger = logging.getLogger(__name__)


class Indexer(object):
    def __init__(self, application_id, api_key, prefix="kinto"):
        self.client = SearchClient.create(application_id, api_key)
        self.prefix = prefix
        self.tasks = []

    def join(self):
        for indexname, taskID in self.tasks:
            index = self.client.init_index(indexname)
            index.wait_task(taskID)
        self.tasks = []

    def set_extra_headers(self, headers):
        self.client._config.headers.update(headers)

    def indexname(self, bucket_id, collection_id):
        return "{}-{}-{}".format(self.prefix, bucket_id, collection_id)

    def create_index(
        self, bucket_id, collection_id, settings=None, wait_for_creation=False
    ):
        if settings is None:
            settings = {}
        self.update_index(
            bucket_id, collection_id, settings=settings, wait_for_task=wait_for_creation
        )

    def update_index(
        self, bucket_id, collection_id, settings=None, wait_for_task=False
    ):
        indexname = self.indexname(bucket_id, collection_id)
        if settings is not None:
            index = self.client.init_index(indexname)
            res = index.set_settings(settings, {"forwardToReplicas": True})
            if wait_for_task:
                res.wait()
            else:
                self.tasks.append((indexname, res[0]["taskID"]))

    def delete_index(self, bucket_id, collection_id=None):
        if collection_id is None:
            response = self.client.list_indices()
            index_prefix = self.indexname(bucket_id, "")
            collections = [
                i["name"]
                for i in response["items"]
                if i["name"].startswith(index_prefix)
            ]
        else:
            collections = [self.indexname(bucket_id, collection_id)]

        for indexname in collections:
            try:
                self.client.init_index(indexname).delete()
            except AlgoliaException as e:  # pragma: no cover
                if "HTTP Code: 404" not in str(e):
                    raise

    def search(self, bucket_id, collection_id, **kwargs):
        indexname = self.indexname(bucket_id, collection_id)
        index = self.client.init_index(indexname)
        query = kwargs.pop("query", "")
        return index.search(query, kwargs)

    def flush(self):
        response = self.client.list_indices()
        for index in response["items"]:
            indexname = index["name"]
            if indexname.startswith(self.prefix):
                index = self.client.init_index(indexname)
                index.clear_objects().wait()
                index.delete().wait()

    def isalive(self):
        self.client._transporter.read(Verb.GET, "1/isalive", {}, None)

    @contextmanager
    def bulk(self):
        bulk = BulkClient(self)
        yield bulk

        for indexname, requests in bulk.operations.items():
            index = self.client.init_index(indexname)
            res = index.batch(requests)
            self.tasks.append((indexname, res[0]["taskID"]))


class BulkClient:
    def __init__(self, indexer):
        self.indexer = indexer
        self.operations = {}

    def index_record(self, bucket_id, collection_id, record, id_field="id"):
        indexname = self.indexer.indexname(bucket_id, collection_id)
        self.operations.setdefault(indexname, [])
        obj = deepcopy(record)
        record_id = obj.pop(id_field)
        obj["objectID"] = record_id
        self.operations[indexname].append({"action": "addObject", "body": obj})

    def unindex_record(self, bucket_id, collection_id, record, id_field="id"):
        indexname = self.indexer.indexname(bucket_id, collection_id)
        record_id = record[id_field]
        self.operations.setdefault(indexname, [])
        self.operations[indexname].append(
            {"action": "deleteObject", "body": {"objectID": record_id}}
        )


def heartbeat(request):
    """Test that Algolia is operationnal.

    :param request: current request object
    :type request: :class:`~pyramid:pyramid.request.Request`
    :returns: ``True`` is everything is ok, ``False`` otherwise.
    :rtype: bool
    """
    indexer = request.registry.indexer
    try:
        indexer.isalive()
    except Exception as e:
        logger.exception(e)
        return False
    else:
        return True


def load_from_config(config):
    settings = config.get_settings()
    application_id = settings.get("algolia.application_id")
    api_key = settings.get("algolia.api_key")
    if application_id is None or api_key is None:
        message = (
            "kinto-algolia needs kinto.algolia.application_id "
            "and kinto.algolia.api_key settings to be set."
        )
        raise ConfigurationError(message)

    prefix = settings.get("algolia.index_prefix", "kinto")
    indexer = Indexer(application_id=application_id, api_key=api_key, prefix=prefix)
    return indexer



import unittest
from unittest import mock

from . import BaseWebTest


class ParentDeletion(BaseWebTest, unittest.TestCase):

    def setUp(self):
        self.app.put("/buckets/bid", headers=self.headers)
        self.app.put("/buckets/bid/collections/cid", headers=self.headers)
        self.app.post_json("/buckets/bid/collections/cid/records",
                           {"data": {"hello": "world"}},
                           headers=self.headers)

    def test_index_is_deleted_when_collection_is_deleted(self):
        with mock.patch.object(self.app.app.registry.indexer, "client") as client:
            self.app.delete("/buckets/bid/collections/cid", headers=self.headers)
        client.init_index.assert_called_once_with('kinto-bid-cid')
        client.init_index.return_value.delete.assert_called_once()

    def test_index_is_deleted_when_bucket_is_deleted(self):
        with mock.patch.object(self.app.app.registry.indexer, "client") as client:
            client.list_indices.return_value = {"items": [
                {"name": "kinto-foo-bar"},
                {"name": "kinto-bid-cid"}
            ]}
            self.app.delete("/buckets/bid", headers=self.headers)
        client.init_index.assert_called_once_with('kinto-bid-cid')
        client.init_index.return_value.delete.assert_called_once()



import pkg_resources

from pyramid.settings import aslist
from kinto.events import ServerFlushed
from kinto.core.events import AfterResourceChanged

from . import indexer
from . import listener


#: Module version, as defined in PEP-0396.
__version__ = pkg_resources.get_distribution(__package__).version


def includeme(config):
    settings = config.get_settings()
    collections = settings.get("algolia.resources", [])
    # Register a global indexer object
    config.registry.indexer = indexer.load_from_config(config)

    # Register heartbeat to check algolia integration.
    config.registry.heartbeats["algolia"] = indexer.heartbeat

    # Activate end-points.
    config.scan("kinto_algolia.views")

    on_record_changed_listener = listener.on_record_changed

    # If StatsD is enabled, monitor execution time of listener.
    if config.registry.statsd:
        statsd_client = config.registry.statsd
        key = "plugins.algolia.index"
        on_record_changed_listener = statsd_client.timer(key)(
            on_record_changed_listener
        )

    config.add_subscriber(
        on_record_changed_listener, AfterResourceChanged, for_resources=("record",)
    )

    config.add_subscriber(listener.on_server_flushed, ServerFlushed)
    config.add_subscriber(
        listener.on_collection_created,
        AfterResourceChanged,
        for_resources=("collection",),
        for_actions=("create",),
    )
    config.add_subscriber(
        listener.on_collection_updated,
        AfterResourceChanged,
        for_resources=("collection",),
        for_actions=("update",),
    )
    config.add_subscriber(
        listener.on_collection_deleted,
        AfterResourceChanged,
        for_resources=("collection",),
        for_actions=("delete",),
    )
    config.add_subscriber(
        listener.on_bucket_deleted,
        AfterResourceChanged,
        for_resources=("bucket",),
        for_actions=("delete",),
    )

    config.add_api_capability(
        "algolia",
        description="Index and search records using Algolia.",
        url="https://github.com/Kinto/kinto-algolia",
        version=__version__,
        collections=aslist(collections),
    )



import logging

from algoliasearch.exceptions import AlgoliaException
from kinto.core.events import ACTIONS
from .utils import is_monitoring_collection


logger = logging.getLogger(__name__)


def on_collection_created(event):
    registry = event.request.registry
    indexer = registry.indexer
    bucket_id = event.payload["bucket_id"]
    for created in event.impacted_records:
        collection_id = created["new"]["id"]
        if is_monitoring_collection(registry, bucket_id, collection_id):
            settings = created["new"].get("algolia:settings")
            indexer.create_index(bucket_id, collection_id, settings=settings)


def on_collection_updated(event):
    registry = event.request.registry
    indexer = registry.indexer
    bucket_id = event.payload["bucket_id"]
    for updated in event.impacted_records:
        collection_id = updated["new"]["id"]

        if is_monitoring_collection(registry, bucket_id, collection_id):
            old_settings = updated["old"].get("algolia:settings")
            new_settings = updated["new"].get("algolia:settings")
            # Create if there was no index before.
            if old_settings != new_settings:
                indexer.update_index(bucket_id, collection_id, settings=new_settings)


def on_collection_deleted(event):
    registry = event.request.registry
    indexer = registry.indexer
    bucket_id = event.payload["bucket_id"]
    for deleted in event.impacted_records:
        collection_id = deleted["old"]["id"]
        if is_monitoring_collection(registry, bucket_id, collection_id):
            indexer.delete_index(bucket_id, collection_id)


def on_bucket_deleted(event):
    registry = event.request.registry
    indexer = registry.indexer
    for deleted in event.impacted_records:
        bucket_id = deleted["old"]["id"]
        if is_monitoring_collection(registry, bucket_id):
            indexer.delete_index(bucket_id)


def on_record_changed(event):
    registry = event.request.registry
    indexer = registry.indexer

    bucket_id = event.payload["bucket_id"]
    collection_id = event.payload["collection_id"]

    if is_monitoring_collection(registry, bucket_id, collection_id):
        action = event.payload["action"]
        try:
            with indexer.bulk() as bulk:
                for change in event.impacted_records:
                    if action == ACTIONS.DELETE.value:
                        bulk.unindex_record(
                            bucket_id, collection_id, record=change["old"]
                        )
                    else:
                        bulk.index_record(
                            bucket_id, collection_id, record=change["new"]
                        )
        except AlgoliaException:
            logger.exception("Failed to index record")


def on_server_flushed(event):
    indexer = event.request.registry.indexer
    indexer.flush()



from pyramid.settings import aslist

from kinto.core import utils as core_utils


def is_monitoring_collection(registry, bucket_id, collection_id=None):
    resources_uri = aslist(registry.settings.get("algolia.resources", ""))

    for resource_uri in resources_uri:
        resource_name, matchdict = core_utils.view_lookup_registry(
            registry, resource_uri
        )
        if (resource_name == "bucket" and bucket_id == matchdict["id"]) or (
            collection_id is None and bucket_id == matchdict["bucket_id"]
        ):
            return True

        is_matching_collection = (
            resource_name == "collection"
            and bucket_id == matchdict["bucket_id"]
            and collection_id == matchdict["id"]
        )
        if is_matching_collection:
            return True



import argparse
import logging
import time
import sys

from algoliasearch.exceptions import AlgoliaException
from pyramid.paster import bootstrap

from kinto.core.storage.exceptions import RecordNotFoundError
from kinto.core.storage import Sort, Filter
from kinto.core.utils import COMPARISON


DEFAULT_CONFIG_FILE = "config/kinto.ini"

logger = logging.getLogger(__package__)


def main(cli_args=None):
    if cli_args is None:
        cli_args = sys.argv[1:]

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--ini",
        help="Application configuration file",
        dest="ini_file",
        required=False,
        default=DEFAULT_CONFIG_FILE,
    )
    parser.add_argument("-b", "--bucket", help="Bucket name.", type=str)
    parser.add_argument("-c", "--collection", help="Collection name.", type=str)
    args = parser.parse_args(args=cli_args)

    print("Load config...")
    env = bootstrap(args.ini_file)
    registry = env["registry"]

    # Make sure that kinto-algolia is configured.
    try:
        indexer = registry.indexer
    except AttributeError:
        logger.error("kinto-algolia not available.")
        return 62

    bucket_id = args.bucket
    collection_id = args.collection

    # Get index settings from collection metadata.
    try:
        settings = get_index_settings(registry.storage, bucket_id, collection_id)
    except RecordNotFoundError:
        logger.error("No collection '%s' in bucket '%s'" % (collection_id, bucket_id))
        return 63

    recreate_index(indexer, bucket_id, collection_id, settings)
    print("Waiting for Algolia quota stats to propagate.")
    for _ in range(3):
        time.sleep(1)  # Wait a couple of seconds
        print(".", end="")
        sys.stdout.flush()
    print()
    reindex_records(indexer, registry.storage, bucket_id, collection_id)

    return 0


def get_index_settings(storage, bucket_id, collection_id):
    # Open collection metadata.
    # XXX: https://github.com/Kinto/kinto/issues/710
    metadata = storage.get(
        parent_id="/buckets/%s" % bucket_id,
        collection_id="collection",
        object_id=collection_id,
    )
    return metadata.get("algolia:settings")


def recreate_index(indexer, bucket_id, collection_id, settings):
    index_name = indexer.indexname(bucket_id, collection_id)
    # Delete existing index.
    indexer.delete_index(bucket_id, collection_id)
    print("Old index '%s' deleted." % index_name)
    # Recreate the index with the new settings.
    indexer.create_index(bucket_id, collection_id, settings=settings)
    print("New index '%s' created." % index_name)


def get_paginated_records(storage, bucket_id, collection_id, limit=5000):
    # We can reach the storage_fetch_limit, so we use pagination.
    parent_id = "/buckets/%s/collections/%s" % (bucket_id, collection_id)
    sorting = [Sort("last_modified", -1)]
    pagination_rules = []
    while "not gone through all pages":
        records, _ = storage.get_all(
            parent_id=parent_id,
            collection_id="record",
            pagination_rules=pagination_rules,
            sorting=sorting,
            limit=limit,
        )
        yield records

        if len(records) < limit:
            break  # Done.

        smallest_timestamp = records[-1]["last_modified"]
        pagination_rules = [
            [Filter("last_modified", smallest_timestamp, COMPARISON.LT)]
        ]


def reindex_records(indexer, storage, bucket_id, collection_id):
    total = 0
    for records in get_paginated_records(storage, bucket_id, collection_id):
        try:
            with indexer.bulk() as bulk:
                for record in records:
                    bulk.index_record(bucket_id, collection_id, record=record)
                print(".", end="")
                sys.stdout.flush()
            total += len(list(bulk.operations.items())[0][1])
        except AlgoliaException:
            logger.exception("Failed to index record")
            break
    print("\n%s records reindexed." % total)



import unittest
from unittest import mock

from algoliasearch.exceptions import AlgoliaException
from pyramid import testing
from pyramid.exceptions import ConfigurationError

import kinto.core

from kinto_algolia import __version__ as algolia_version, includeme
from . import BaseWebTest
from kinto import main


class PluginSetup(BaseWebTest, unittest.TestCase):

    def test_algolia_capability_exposed(self):
        resp = self.app.get('/')
        capabilities = resp.json['capabilities']
        self.assertIn('algolia', capabilities)
        expected = {
            "version": algolia_version,
            "description": "Index and search records using Algolia.",
            "url": "https://github.com/Kinto/kinto-algolia",
            "collections": ["/buckets/bid/collections/cid"],
        }
        assert expected == capabilities['algolia']

    def test_indexer_flush(self):
        with mock.patch("kinto_algolia.indexer.Indexer.flush") as flush:
            self.app.post("/__flush__", status=202)
            assert flush.called

    def test_present_in_heartbeat(self):
        resp = self.app.get("/__heartbeat__")
        assert "algolia" in resp.json

    def test_returns_false_if_connection_fails(self):
        with mock.patch.object(self.app.app.registry.indexer, "client") as client:
            client._transporter.read.side_effect = AlgoliaException
            resp = self.app.get("/__heartbeat__", status=503)
            assert not resp.json["algolia"]

    def test_include_fails_if_missing_config(self):
        config = testing.setUp()
        settings = config.get_settings()
        kinto.core.initialize(config, '0.0.1')
        with self.assertRaises(ConfigurationError) as e:
            config.include(includeme)
            main({}, None, **settings)
        assert str(e.exception) == ('kinto-algolia needs kinto.algolia.application_id '
                                    'and kinto.algolia.api_key settings to be set.')



import unittest
from unittest import mock

from algoliasearch.exceptions import AlgoliaException
from kinto.core.testing import get_user_headers

from . import BaseWebTest


class SearchView(BaseWebTest, unittest.TestCase):
    def setUp(self):
        self.app.put("/buckets/bid", headers=self.headers)
        self.app.put("/buckets/bid/collections/cid", headers=self.headers)

    def test_search_response_error_400_indexer_fails(self):
        with mock.patch.object(self.app.app.registry.indexer, "client") as client:
            client.init_index.return_value.search.side_effect = AlgoliaException
            self.app.post("/buckets/bid/collections/cid/search", headers=self.headers, status=400)

    def test_search_response_error_400_with_wrong_body(self):
        self.app.post("/buckets/bid/collections/cid/search", 'blah',
                      headers=self.headers, status=400)

    def test_invalid_search_query(self):
        body = {"whatever": {"wrong": "bad"}}
        resp = self.app.post_json("/buckets/bid/collections/cid/search",
                                  body,
                                  headers=self.headers,
                                  status=400)
        assert resp.json["message"] == "Algolia error in body: Unknown parameter: whatever"

    def test_search_on_empty_collection_returns_empty_list(self):
        resp = self.app.post("/buckets/bid/collections/cid/search",
                             headers=self.headers)
        result = resp.json
        assert len(result["hits"]) == 0

    def test_querystring_search_is_supported(self):
        self.app.post_json("/buckets/bid/collections/cid/records",
                           {"data": {"age": 12}}, headers=self.headers)
        self.app.post_json("/buckets/bid/collections/cid/records",
                           {"data": {"age": 21}}, headers=self.headers)
        self.indexer.join()
        resp = self.app.get("/buckets/bid/collections/cid/search?filters=age<15",
                            headers=self.headers)
        result = resp.json
        assert len(result["hits"]) == 1
        assert result["hits"][0]["age"] == 12

    def test_empty_querystring_returns_all_results(self):
        self.app.post_json("/buckets/bid/collections/cid/records",
                           {"data": {"age": 12}}, headers=self.headers)
        self.app.post_json("/buckets/bid/collections/cid/records",
                           {"data": {"age": 21}}, headers=self.headers)
        self.indexer.join()
        resp = self.app.get("/buckets/bid/collections/cid/search",
                            headers=self.headers)
        result = resp.json
        assert len(result["hits"]) == 2


# ALGOLIA SEARCH DOESN'T SUPPORT LIMITING YET
# https://github.com/algolia/algoliasearch-client-python/issues/365
# class LimitedResults(BaseWebTest, unittest.TestCase):
#     def get_app(self, settings):
#         app = self.make_app(settings=settings)
#         app.put("/buckets/bid", headers=self.headers)
#         app.put_json("/buckets/bid/collections/cid",
#                      {"data": {"algolia:settings": {}}},
#                      headers=self.headers)
#         requests = [{
#             "method": "POST",
#             "path": "/buckets/bid/collections/cid/records",
#             "body": {"data": {"age": i}}
#         } for i in range(5)]
#         app.post_json("/batch", {"requests": requests}, headers=self.headers)
#         sleep(1)  # Wait for indexing
#         return app
#
#     def test_the_number_of_responses_is_limited_by_paginate_by_setting(self):
#         app = self.get_app({"paginate_by": 2})
#         resp = app.get("/buckets/bid/collections/cid/search", headers=self.headers)
#         result = resp.json
#         assert len(result["hits"]) == 2
#
#     def test_the_number_of_responses_is_limited_by_max_fetch_size_setting(self):
#         app = self.get_app({"storage_max_fetch_size": 2})
#         resp = app.get("/buckets/bid/collections/cid/search", headers=self.headers)
#         result = resp.json
#         assert len(result["hits"]) == 2
#
#     def test_the_number_of_responses_is_limited_by_smaller_limit(self):
#         app = self.get_app({"paginate_by": 4, "storage_max_fetch_size": 2})
#         resp = app.get("/buckets/bid/collections/cid/search", headers=self.headers)
#         result = resp.json
#         assert len(result["hits"]) == 2
#
#     def test_the_number_of_responses_is_limited_by_only_defined_limit(self):
#         app = self.get_app({"paginate_by": 0, "storage_max_fetch_size": 2})
#         resp = app.get("/buckets/bid/collections/cid/search", headers=self.headers)
#         result = resp.json
#         assert len(result["hits"]) == 2
#
#     def test_size_specified_in_query_is_taken_into_account(self):
#         app = self.get_app({"paginate_by": 3})
#         query = {
#             "hitsPerPage": 2
#         }
#         resp = app.post_json("/buckets/bid/collections/cid/search", query,
#                              headers=self.headers)
#         result = resp.json
#         assert len(result["hits"]) == 2
#
#     def test_size_specified_in_query_is_caped_by_setting(self):
#         app = self.get_app({"paginate_by": 3})
#         query = {
#             "hitsPerPage": 4
#         }
#         resp = app.post_json("/buckets/bid/collections/cid/search", query,
#                              headers=self.headers)
#         result = resp.json
#         assert len(result["hits"]) == 3


class PermissionsCheck(BaseWebTest, unittest.TestCase):
    def test_search_is_allowed_if_write_on_bucket(self):
        body = {"permissions": {"write": ["system.Everyone"]}}
        self.app.put_json("/buckets/bid", body, headers=self.headers)
        self.app.put("/buckets/bid/collections/cid", headers=self.headers)

        self.app.post("/buckets/bid/collections/cid/search", status=200)

    def test_search_is_allowed_if_read_on_bucket(self):
        body = {"permissions": {"read": ["system.Everyone"]}}
        self.app.put_json("/buckets/bid", body, headers=self.headers)
        self.app.put("/buckets/bid/collections/cid", headers=self.headers)

        self.app.post("/buckets/bid/collections/cid/search", status=200)

    def test_search_is_allowed_if_write_on_collection(self):
        self.app.put("/buckets/bid", headers=self.headers)
        body = {"permissions": {"write": ["system.Everyone"]}}
        self.app.put_json("/buckets/bid/collections/cid", body, headers=self.headers)

        self.app.post("/buckets/bid/collections/cid/search", status=200)

    def test_search_is_allowed_if_read_on_collection(self):
        self.app.put("/buckets/bid", headers=self.headers)
        body = {"permissions": {"read": ["system.Everyone"]}}
        self.app.put_json("/buckets/bid/collections/cid", body, headers=self.headers)

        self.app.post("/buckets/bid/collections/cid/search", status=200)

    def test_search_is_not_allowed_by_default(self):
        self.app.put("/buckets/bid", headers=self.headers)
        self.app.put("/buckets/bid/collections/cid", headers=self.headers)

        self.app.post("/buckets/bid/collections/cid/search", status=401)
        headers = get_user_headers("cual", "quiera")
        self.app.post("/buckets/bid/collections/cid/search", status=403, headers=headers)

    def test_search_is_not_allowed_if_only_read_on_certain_records(self):
        self.app.put("/buckets/bid", headers=self.headers)
        body = {"permissions": {"record:create": ["system.Authenticated"]}}
        self.app.put_json("/buckets/bid/collections/cid", body, headers=self.headers)
        headers = get_user_headers("toto")
        self.app.post_json("/buckets/bid/collections/cid/records", {"data": {"pi": 42}},
                           headers=headers)

        self.app.post("/buckets/bid/collections/cid/search", status=403, headers=headers)



import unittest
from unittest import mock

from algoliasearch.exceptions import AlgoliaException

from . import BaseWebTest


class RecordIndexing(BaseWebTest, unittest.TestCase):

    def setUp(self):
        self.app.put("/buckets/bid", headers=self.headers)
        self.app.put("/buckets/bid/collections/cid", headers=self.headers)
        resp = self.app.post_json("/buckets/bid/collections/cid/records",
                                  {"data": {"hello": "world"}},
                                  headers=self.headers)
        self.record = resp.json["data"]
        self.indexer.join()

    def test_new_index_settings_are_updated(self):
        self.app.put_json("/buckets/bid/collections/cid",
                          {"data": {"algolia:settings": {}}},
                          headers=self.headers)

    def test_new_records_are_indexed(self):
        resp = self.app.post("/buckets/bid/collections/cid/search",
                             headers=self.headers)
        result = resp.json
        assert len(result["hits"]) == 1
        hit = result["hits"][0]
        del hit['_highlightResult']
        hit['id'] = hit.pop('objectID')
        assert hit == self.record

    def test_deleted_records_are_unindexed(self):
        rid = self.record["id"]
        self.app.delete("/buckets/bid/collections/cid/records/{}".format(rid),
                        headers=self.headers)
        self.indexer.join()

        resp = self.app.post("/buckets/bid/collections/cid/search",
                             headers=self.headers)
        result = resp.json
        assert len(result["hits"]) == 0

    def test_response_is_served_if_indexer_fails(self):
        with mock.patch.object(self.app.app.registry.indexer, "client") as client:
            client.init_index.return_value.batch.side_effect = AlgoliaException
            r = self.app.post_json("/buckets/bid/collections/cid/records",
                                   {"data": {"hola": "mundo"}},
                                   headers=self.headers)
            assert r.status_code == 201

    def test_a_statsd_timer_is_used_if_configured(self):
        statsd_client = self.app.app.registry.statsd._client
        with mock.patch.object(statsd_client, 'timing') as mocked:
            self.app.post_json("/buckets/bid/collections/cid/records",
                               {"data": {"hola": "mundo"}},
                               headers=self.headers)
            timers = set(c[0][0] for c in mocked.call_args_list)
            assert 'plugins.algolia.index' in timers



import os
import unittest
from unittest import mock

from algoliasearch.exceptions import AlgoliaException
from kinto_algolia.command_reindex import main, reindex_records, get_paginated_records

from . import BaseWebTest

HERE = os.path.abspath(os.path.dirname(__file__))


class TestMain(BaseWebTest, unittest.TestCase):

    schema = {
        "attributesToIndex": ["name"]
    }

    def test_cli_fail_if_algolia_plugin_not_installed(self):
        with mock.patch('kinto_algolia.command_reindex.logger') as logger:
            exit_code = main(['--ini', os.path.join(HERE, 'wrong_config.ini'),
                              '--bucket', 'bid', '--collection', 'cid'])
            assert exit_code == 62
            logger.error.assert_called_with('kinto-algolia not available.')

    def test_cli_fail_if_collection_or_bucket_do_not_exists(self):
        with mock.patch('kinto_algolia.command_reindex.logger') as logger:
            exit_code = main(['--ini', self.ini_path(),
                              '--bucket', 'bid', '--collection', 'cid'])
            assert exit_code == 63
            logger.error.assert_called_with(
                "No collection 'cid' in bucket 'bid'")

    def test_cli_reindexes_if_collection_has_an_index_schema(self):
        # Create collection or bucket
        self.app.put("/buckets/bid", headers=self.headers)
        body = {"data": {"algolia:settings": self.schema}}
        self.app.put_json("/buckets/bid/collections/cid", body, headers=self.headers)
        self.app.post_json("/buckets/bid/collections/cid/records",
                           {"data": {"build": {"id": "efg", "date": "2017-02-01"}}},
                           headers=self.headers)

        exit_code = main(['--ini', self.ini_path(),
                          '--bucket', 'bid', '--collection', 'cid'])
        assert exit_code == 0

    def test_cli_logs_algolia_exceptions(self):
        indexer = mock.MagicMock()
        indexer.bulk().__enter__().index_record.side_effect = AlgoliaException

        with mock.patch('kinto_algolia.command_reindex.logger') as logger:
            with mock.patch('kinto_algolia.command_reindex.get_paginated_records',
                            return_value=[[{}, {}]]) as get_paginated_records:
                reindex_records(indexer,
                                mock.sentinel.storage,
                                mock.sentinel.bucket_id,
                                mock.sentinel.collection_id)
                get_paginated_records.assert_called_with(mock.sentinel.storage,
                                                         mock.sentinel.bucket_id,
                                                         mock.sentinel.collection_id)
                logger.exception.assert_called_with('Failed to index record')

    def test_cli_default_to_sys_argv(self):
        with mock.patch('sys.argv', ['cli', '--ini', os.path.join(HERE, 'wrong_config.ini')]):
            exit_code = main()
            assert exit_code == 62

    def test_get_paginated_records(self):
        # Create collection or bucket
        self.app.put("/buckets/bid", headers=self.headers)
        body = {"data": {"algolia:settings": self.schema}}
        self.app.put_json("/buckets/bid/collections/cid", body, headers=self.headers)
        for i in range(5):
            self.app.post_json("/buckets/bid/collections/cid/records",
                               {"data": {"build": {"id": "efg%d" % i, "date": "2017-02-01"}}},
                               headers=self.headers)
        count = 0
        for records in get_paginated_records(self.app.app.registry.storage, 'bid', 'cid', limit=3):
            count += 1
        assert count == 2



import unittest

from . import BaseWebTest


class PostActivation(BaseWebTest, unittest.TestCase):

    def setUp(self):
        app = self.make_app(settings={"kinto.includes": ""})
        capabilities = app.get("/").json["capabilities"]
        assert "algolia" not in capabilities

        app.put("/buckets/bid", headers=self.headers)
        app.put("/buckets/bid/collections/cid", headers=self.headers)
        app.post_json("/buckets/bid/collections/cid/records",
                      {"data": {"before": "indexing"}},
                      headers=self.headers)

    def test_search_does_not_fail(self):
        resp = self.app.get("/buckets/bid/collections/cid/records",
                            headers=self.headers)
        assert len(resp.json["data"]) == 1

        resp = self.app.get("/buckets/bid/collections/cid/search",
                            headers=self.headers)
        results = resp.json
        assert len(results["hits"]) == 0

    def test_record_creation_does_not_fail(self):
        self.app.post_json("/buckets/bid/collections/cid/records",
                           {"data": {"after": "indexing"}},
                           headers=self.headers)
        self.indexer.join()
        resp = self.app.get("/buckets/bid/collections/cid/search",
                            headers=self.headers)
        results = resp.json
        assert len(results["hits"]) == 1
        assert results["hits"][0]["after"] == "indexing"



#!/usr/bin/env python
from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.rst') as history_file:
    history = history_file.read()

REQUIREMENTS = [
    'algoliasearch',
    'kinto'
]

TEST_REQUIREMENTS = [
    'webtest',
]

ENTRY_POINTS = {
    'console_scripts': [
        'kinto-algolia-reindex = kinto_algolia.command_reindex:main'
    ],
}

setup(
    name='kinto-algolia',
    version='1.1.0',
    description="Index and search records using Algolia.",
    long_description=readme + '\n\n' + history,
    author='Rémy Hubscher',
    author_email='hubscher.remy@gmail.com',
    url='https://github.com/Kinto/kinto-algolia',
    packages=[
        'kinto_algolia',
    ],
    package_dir={'kinto_algolia': 'kinto_algolia'},
    include_package_data=True,
    install_requires=REQUIREMENTS,
    license="Apache License (2.0)",
    zip_safe=False,
    keywords='kinto algolia index',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=TEST_REQUIREMENTS,
    entry_points=ENTRY_POINTS
)



import json
import logging

from algoliasearch.exceptions import AlgoliaException
from kinto.core import authorization
from kinto.core import Service
from kinto.core import utils
from kinto.core.errors import raise_invalid


logger = logging.getLogger(__name__)


class RouteFactory(authorization.RouteFactory):
    def __init__(self, request):
        super().__init__(request)
        records_plural = utils.strip_uri_prefix(
            request.path.replace("/search", "/records")
        )
        self.permission_object_id = records_plural
        self.required_permission = "read"


search = Service(
    name="search",
    path="/buckets/{bucket_id}/collections/{collection_id}/search",
    description="Search",
    factory=RouteFactory,
)


def search_view(request, **kwargs):
    bucket_id = request.matchdict["bucket_id"]
    collection_id = request.matchdict["collection_id"]

    # algoliasearch doesn't support pagination
    # https://github.com/algolia/algoliasearch-client-python/issues/365
    #
    # Limit the number of results to return, based on existing Kinto settings.
    # paginate_by = request.registry.settings.get("paginate_by")
    # max_fetch_size = request.registry.settings["storage_max_fetch_size"]
    # if paginate_by is None or paginate_by <= 0:
    #     paginate_by = max_fetch_size
    # configured = min(paginate_by, max_fetch_size)
    # # If the size is specified in query, ignore it if larger than setting.
    # specified = None
    # if "size" in kwargs:
    #     specified = kwargs.get("hitsPerPage")
    #
    # if specified is None or specified > configured:
    #     kwargs.setdefault("hitsPerPage", configured)

    # Access indexer from views using registry.
    indexer = request.registry.indexer
    try:
        indexer.set_extra_headers(
            {"Referer": request.headers.get("Referer", request.route_url("hello"))}
        )
        results = indexer.search(bucket_id, collection_id, **kwargs)
    except AlgoliaException as e:
        logger.exception("Index query failed.")
        message = str(e)
        if "does not exist" in message:
            # If plugin was enabled after the creation of the collection.
            indexer.create_index(bucket_id, collection_id, wait_for_creation=True)
            return search_view(request, **kwargs)
        else:
            error_details = {"name": "Algolia error", "description": message}
            return raise_invalid(request, **error_details)

    return results


@search.post(permission=authorization.DYNAMIC)
def post_search(request):
    try:
        body = json.loads(request.body.decode("utf-8"))
    except json.decoder.JSONDecodeError:
        if not request.body:
            body = {}
        else:
            error_details = {
                "name": "JSONDecodeError",
                "description": "Please make sure your request body is a valid JSON payload.",
            }
            raise_invalid(request, **error_details)

    return search_view(request, **body)


@search.get(permission=authorization.DYNAMIC)
def get_search(request):
    kwargs = dict(**request.GET)
    return search_view(request, **kwargs)
